import logging
from typing import Any, Dict, Iterator, List

from confluent_kafka import Consumer
from flask import current_app

from dh_potluck.messaging import Message, MessageEnvelopeSchema

LOG = logging.getLogger(__name__)


class MessageConsumer(object):

    _consumer: Consumer
    _is_consuming: bool = False
    _message_schema: MessageEnvelopeSchema = MessageEnvelopeSchema()

    def __init__(
        self,
        consumer_group: str,
        config_overrides: Dict[str, Any] = None,
    ):
        """
        :param str consumer_group: The kafka consumer group to use
        :param dict config_overrides: Any kafka configuration overrides
        """
        brokers: str = current_app.config['KAFKA_BROKERS_LIST']
        if brokers is None:
            raise RuntimeError(
                'Tried to instantiate a MessageConsumer '
                'without setting the KAFKA_BROKERS_LIST env var'
            )
        config = {
            'bootstrap.servers': brokers,
            'group.id': consumer_group,
            'auto.offset.reset': 'earliest',
            'enable.auto.commit': 'false',
        }
        should_connect_ssl: bool = current_app.config['KAFKA_USE_SSL_CONNECTION']
        if should_connect_ssl:
            config['security.protocol'] = 'SSL'
        if config_overrides:
            config.update(config_overrides)
        self._consumer = Consumer(config)

    def subscribe(self, topics: List[str]) -> None:
        """
        Subscribe to the list of topics, this makes the connection to Kafka
        :param topics: Topics to connect to
        :return: None
        """
        self._consumer.subscribe(topics)

    def get_messages(self, poll_timeout: float = 1.0) -> Iterator[Message]:
        """
        Get messages from topics

        :param float poll_timeout: Maximum time to block waiting for message. (Seconds)
        :return: Iterator[Message]
        """
        self._is_consuming = True
        while self._is_consuming:
            message = self._consumer.poll(poll_timeout)

            if message is None:
                continue
            if message.error():
                LOG.error(f'Consumer error: {message.error()}')
                continue

            LOG.debug(f'Received message: {message.value().decode("utf-8")}')
            message_value = self._message_schema.loads(message.value().decode('utf-8'))
            yield Message(message.topic(), message_value, message)

    def commit(self, message: Message) -> None:
        """
        Commit the message, this is done after successfully processing the message

        :param Message message: Message to commit
        :return: None
        """
        self._consumer.commit(message.raw_message())

    def shutdown(self) -> None:
        """
        Call when service is shutting down

        :return: None
        """
        LOG.info('Consumer shutting down...')
        self._is_consuming = False
        self._consumer.close()
        LOG.info('Consumer shut down')

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()



from typing import Any, Dict, Protocol

from ddtrace import tracer
from marshmallow import Schema

from dh_potluck.messaging import Message, MessageConsumer


class MessageHandlerCallback(Protocol):
    def __call__(self, message_type: str, payload: Dict) -> bool:
        pass


class MessageHandler(object):
    _schema: Schema
    _handler: MessageHandlerCallback

    def __init__(self, schema: Schema, handler: MessageHandlerCallback):
        """
        :param Schema schema: marshmallow schema to use when deserializing the message
        :param MessageHandlerCallback handler: Function to handle each message. Return True to
                                               commit the message, False to continue without
                                               committing. Raise an error if you are unable to
                                               handle the message.
        """
        self._schema = schema
        self._handler = handler

    def deserialize_message(self, message: Message) -> Dict:
        return self._schema.loads(message.value()['payload'])

    def handle(self, message_type: str, payload: Dict) -> bool:
        should_commit = self._handler(message_type, payload)
        if should_commit is None:
            return True
        return should_commit


class NoHandlerException(Exception):
    def __init__(self, topic):
        self.topic = topic
        self.message = f'No handler for topic: {topic}'
        super().__init__(self.message)


class IncomingMessageRouter(object):
    _handlers: Dict[str, MessageHandler]
    _consumer: MessageConsumer

    def __init__(
        self,
        handlers: Dict[str, MessageHandler],
        consumer_group: str,
        config_overrides: Dict[str, Any] = None,
    ):
        """
        :param dict handlers: Dictionary of topic -> MessageHandler
        :param str consumer_group: The kafka consumer group to use
        :param dict config_overrides: Any kafka configuration overrides
        """
        self._consumer = MessageConsumer(consumer_group, config_overrides)
        self._handlers = handlers

    def run(self) -> None:
        """
        Start consuming messages. On new messages, check the handlers map, if the message's topic
        matches a handler key, use it to serialize the message, and handle it.
        :return: None
        """
        self._consumer.subscribe(list(self._handlers.keys()))
        try:
            for message in self._consumer.get_messages():
                self._handle_message(message)
        finally:
            self._consumer.shutdown()

    def _handle_message(self, message: Message) -> None:
        topic = message.topic()
        handler = self._handlers.get(topic)
        if not handler:
            raise NoHandlerException(topic)
        payload = handler.deserialize_message(message)
        message_type = message.value()['message_type']
        with tracer.trace('kafka.consume', resource=f'{topic} {message_type}'):
            should_commit = handler.handle(message_type, payload)
        if should_commit:
            self._consumer.commit(message)



from .auth import ApplicationUser, UnauthenticatedUser
from .celery import signals as CelerySignals
from .celery.crontab_parser import parse_crontab_str
from .celery.synchronization import Semaphore, SemaphoreLocked
from .decorators import retry
from .email import EmailService
from .environment import Environment
from .extension import DHPotluck
from .fields import EnumField, UTCDateTime
from .health_checks import HealthChecks
from .image_api import ImageApi
from .messaging import (
    IncomingMessageRouter,
    Message,
    MessageConsumer,
    MessageConsumerApp,
    MessageEnvelopeSchema,
    MessageHandler,
    MessageHandlerCallback,
    MessageProducer,
    message_handler,
)
from .mixpanel import MixpanelService
from .platform_connection import (
    BadApiResponse,
    InvalidPlatformConnection,
    MissingPlatformConnection,
    PlatformConnection,
)
from .s3_service import S3Service

__all__ = [
    'DHPotluck',
    'EnumField',
    'UTCDateTime',
    'ApplicationUser',
    'UnauthenticatedUser',
    'PlatformConnection',
    'BadApiResponse',
    'MissingPlatformConnection',
    'InvalidPlatformConnection',
    'CelerySignals',
    'HealthChecks',
    'retry',
    'Environment',
    'ImageApi',
    'IncomingMessageRouter',
    'Message',
    'MessageConsumer',
    'MessageConsumerApp',
    'MessageEnvelopeSchema',
    'MessageHandler',
    'MessageHandlerCallback',
    'MessageProducer',
    'message_handler',
    'MixpanelService',
    'EmailService',
    'S3Service',
    'Semaphore',
    'SemaphoreLocked',
    'parse_crontab_str',
]



from typing import Dict

from confluent_kafka.cimpl import Message as KafkaMessage


class Message(object):

    _topic: str
    _value: Dict
    _raw_message: KafkaMessage

    def __init__(self, topic: str, value: Dict, raw_message: KafkaMessage):
        """
        :param topic: The topic the message came from
        :param value: The message value
                      {'message_type': str, 'message_time': DateTime, 'payload': str}
        :param raw_message: The raw kafka message object
        """
        self._topic = topic
        self._value = value
        self._raw_message = raw_message

    def topic(self) -> str:
        return self._topic

    def value(self) -> Dict:
        return self._value

    def raw_message(self) -> KafkaMessage:
        return self._raw_message



import mimetypes
import os
import secrets

import boto3
from flask import current_app

SECONDS_IN_A_YEAR = 3600 * 24 * 365


class S3Service:
    _s3 = None

    @classmethod
    def _get_s3(cls):
        if not cls._s3:
            session = boto3.session.Session(
                aws_access_key_id=current_app.config['DH_POTLUCK_AWS_ACCESS_KEY_ID'],
                aws_secret_access_key=current_app.config['DH_POTLUCK_AWS_SECRET_ACCESS_KEY'],
            )
            cls._s3 = session.resource('s3')
        return cls._s3

    @classmethod
    def upload_file(
        cls,
        local_file_uri: str,
        bucket_name: str,
        s3_file_path: str,
        extra_args: dict = None,
        add_secret_to_filename: bool = True,
    ) -> str:
        """
        upload file to s3 bucket and return public s3 url
        @param local_file_uri: full path of file being uploaded
        @param bucket_name: name of bucket to upload to on s3
        @param s3_file_path: full target path of s3 file to upload to
        @param extra_args: additional params such as ContentType and CacheControl
        @param add_secret_to_filename: if to add a secret to the filename. Prehashed filenames
        can override this parameter
        """
        try:
            s3 = cls._get_s3()
            bucket = s3.Bucket(bucket_name)

            if not extra_args:
                type, encoding = mimetypes.guess_type(local_file_uri)
                type = type or 'application/octet-stream'
                extra_args = {
                    'ContentType': type,
                    'ACL': 'public-read',
                    'CacheControl': 'max-age %d' % SECONDS_IN_A_YEAR,
                }

            if add_secret_to_filename:
                s3_file_path = cls.add_secret_to_filename(s3_file_path)
            bucket.upload_file(local_file_uri, s3_file_path, ExtraArgs=extra_args)
            return f'https://{bucket_name}.s3.amazonaws.com/{s3_file_path}'
        except Exception as e:
            raise S3ServiceException(message=f'Error occurred while uploading file to S3: {str(e)}')

    @staticmethod
    def add_secret_to_filename(file_uri: str) -> str:
        name, ext = os.path.splitext(file_uri)
        secret = secrets.token_hex(16)
        return f'{name}-{secret}{ext}'


class S3ServiceException(Exception):
    def __init__(self, message: str = 'Error occurred while uploading file to S3.') -> None:
        self.message = message
        super().__init__(self.message)



from .message import Message  # isort:skip # NOQA
from .message_envelope_schema import MessageEnvelopeSchema  # isort:skip # NOQA
from .consumer import MessageConsumer  # isort:skip # NOQA
from .incoming_message_router import (  # isort:skip # NOQA
    IncomingMessageRouter,
    MessageHandler,
    MessageHandlerCallback,
)
from .producer import MessageProducer  # isort:skip # NOQA
from .message_consumer_app import MessageConsumerApp, message_handler  # isort:skip  # NOQA



from typing import Dict


def parse_crontab_str(crontab_str: str) -> Dict[str, str]:
    """
    Parse a crontab expression and convert to dict that can be passed as kwargs to crontab from
    celery.schedules

    See https://crontab.guru/ for help with generating crontab strings
    """
    try:
        m, h, dw, dm, my = crontab_str.split()
    except ValueError:
        raise ValueError(
            f'Crontab string: {crontab_str} does not contain 5 values, '
            'enter a valid crontab input in the form of "* * * * *"'
        )
    return {'minute': m, 'hour': h, 'day_of_week': dw, 'day_of_month': dm, 'month_of_year': my}
